/*
 * MVCOMP compressor
 * Copyright (C) 2024 Mateusz Viste
 *
 * MIT License.
 *
 * This is a simple example implementation, not a production-ready tool.
 *
 * https://mvcomp.sourceforge.io
 */


#include <stdio.h>
#include <string.h>


/* write qlen literal bytes into dst, returns amount of "compressed" bytes */
static unsigned short mvcomp_litqueue_dump(unsigned short **dst, const unsigned char *q, unsigned short qlen) {
  unsigned short complen = 0;

  AGAIN:

  /* are we done? (also take care of guys calling me in for jokes) */
  if (qlen == 0) return(complen);

  qlen--; /* now it's between 0 and 30 */
  /* write the length and first char */
  **dst = (unsigned short)((qlen / 2) << 8) | q[0];
  *dst += 1;
  q++;
  complen += 2;

  /* anything left? */
  if (qlen == 0) return(complen);

  /* write the pending words */
  if (qlen > 1) {
    memcpy(*dst, q, (qlen/2)*2);
    *dst += qlen / 2;
    q += (qlen / 2) * 2;
    complen += (qlen / 2) * 2;
    qlen -= (qlen / 2) * 2;
  }

  /* one byte might still be left if it did not fit inside a word */
  goto AGAIN;
}


/* compare up to n bytes of locations s1 and s2, returns the amount of same bytes (0..n) */
static unsigned short comparemem(const unsigned char *s1, const unsigned char *s2, unsigned short n) {
  unsigned short i;
  for (i = 0; (i < n) && (s1[i] == s2[i]); i++);
  return(i);
}


/* mvcomp applies the MV-COMPRESSION algorithm to data and returns the compressed size
 * updates len with the number of input bytes left unprocessed */
static unsigned short mvcomp(void *dstbuf, size_t dstbufsz, const unsigned char *src, size_t *len) {
  unsigned short complen = 0;
  unsigned short *dst = dstbuf;
  static unsigned short bytesprocessed = 0;
  unsigned char litqueue[32];
  unsigned char litqueuelen = 0;

  /* read src byte by byte, len times, each time look for a match of 15,14,13..2 chars in the back buffer */
  while (*len > 0) {
    unsigned short matchlen;
    unsigned short minmatch;
    unsigned short offset;
    matchlen = 16;
    if (*len < matchlen) matchlen = (unsigned short)(*len);

    /* abort if no space in output buffer, but do NOT break a literal queue */
    if ((complen >= dstbufsz - 32) && (litqueuelen == 0)) return(complen);

    /* look for a minimum match of 2 bytes, unless I have some pending literal bytes
     * awaiting, in which case I am going through a new data pattern and it is more
     * efficient to wait for a longer match before breaking the literal string */
    if (litqueuelen & 1) {
      minmatch = 3; /* breaking an uneven queue is less expensive */
    } else if (litqueuelen > 0) {
      goto NOMATCH; /* breaking an even-sized literal queue is never a good idea */
    } else {
      minmatch = 2;
    }

    if (matchlen >= minmatch) {
      /* start at -1 and try to match something moving backward. note that
       * matching a string longer than the offset is perfectly valid, this
       * allows for encoding self-duplicating strings (see MVCOMP.TXT) */
      unsigned short maxoffset = 4096;
      unsigned short longestmatch = 0;
      unsigned short longestmatchoffset = 0;
      if (maxoffset > bytesprocessed) maxoffset = bytesprocessed;

      for (offset = 1; offset <= maxoffset; offset++) {
        unsigned short matchingbytes;
        /* quick skip if first two bytes to not match (never interested in 1-byte matches) */
        if (*((const unsigned short *)src) != *(const unsigned short *)(src - offset)) continue;
        /* compute the exact number of bytes that match */
        matchingbytes = comparemem(src, src - offset, matchlen);
        if (matchingbytes == matchlen) {
          //printf("Found match of %u bytes at offset -%u: '%c%c%c...'\n", matchlen, offset, src[0], src[1], src[2]);
          goto FOUND;
        }
        if (matchingbytes > longestmatch) {
          longestmatch = matchingbytes;
          longestmatchoffset = offset ;
        }
      }
      /* is the longest match interesting? */
      if (longestmatch >= minmatch) {
        matchlen = longestmatch;
        offset = longestmatchoffset;
        goto FOUND;
      }
    }

    NOMATCH:

    /* if here: no match found, write a literal byte to queue */
    litqueue[litqueuelen++] = *src;
    src++;
    bytesprocessed++;
    *len -= 1;

    /* dump literal queue to dst if max length reached */
    if (litqueuelen == 31) {
      complen += mvcomp_litqueue_dump(&dst, litqueue, litqueuelen);
      litqueuelen = 0;
    }
    continue;

    FOUND: /* found a match of matchlen bytes at -offset */

    /* dump awaiting literal queue to dst first */
    if (litqueuelen != 0) {
      complen += mvcomp_litqueue_dump(&dst, litqueue, litqueuelen);
      litqueuelen = 0;
    }

    *dst = (unsigned short)((matchlen - 1) << 12) | (offset - 1);
    dst++;
    src += matchlen;
    bytesprocessed += matchlen;
    *len -= matchlen;
    complen += 2;
  }

  /* dump awaiting literal queue to dst first */
  if (litqueuelen != 0) {
    complen += mvcomp_litqueue_dump(&dst, litqueue, litqueuelen);
    litqueuelen = 0;
  }

  return(complen);
}


int main(int argc, char **argv) {
#define BUFOUTSZ 128
#define MAX_UNCOMP_CHUNK (BUFOUTSZ * 16)
#define BUFINSZ (4096 + MAX_UNCOMP_CHUNK * 2)
  FILE *fdin, *fdout;
  static unsigned char bufin[BUFINSZ];
  unsigned char bufout[BUFOUTSZ];
  size_t bufin_len = 0, bufout_len;
  size_t processedinput = 0;
  size_t totread = 0, totwrite = 0;

  /* help screen? */
  if (argc != 3) {
    puts("usage: mvcomp file_in file_out");
    return(1);
  }

  /* make sure output file does not exist */
  fdin = fopen(argv[2], "rb");
  if (fdin != NULL) {
    fclose(fdin);
    puts("ERROR: output file already exists");
    return(1);
  }

  /* open the input file */
  fdin = fopen(argv[1], "rb");
  if (fdin == NULL) {
    puts("ERROR: failed to open input file");
    return(1);
  }

  /* open the output file */
  fdout = fopen(argv[2], "wb");
  if (fdout == NULL) {
    puts("ERROR: failed to open output file");
    return(1);
  }

  /* compress */
  do {

    /* fill the input buffer if it's too low */
    if (bufin_len < MAX_UNCOMP_CHUNK) {
      size_t l = fread(bufin + processedinput + bufin_len, 1, MAX_UNCOMP_CHUNK, fdin);
      bufin_len += l;
      totread += l;
    }

    /* let mvcomp do some work */
    {
      size_t inputlen = bufin_len;
      bufout_len = mvcomp(bufout, BUFOUTSZ, bufin + processedinput, &bufin_len);
      /* figure out how many bytes of input have been processed */
      processedinput += inputlen - bufin_len;
    }

    /* rewind the input buffer when it reaches the end of the buffer, but
     * always leave a 4K backreference window */
    if (processedinput > BUFINSZ - (MAX_UNCOMP_CHUNK * 2)) {
      memmove(bufin, bufin + processedinput - 4096, BUFINSZ - (processedinput - 4096));
      processedinput = 4096;
    }

    /* write to output */
    if (fwrite(bufout, 1, bufout_len, fdout) != bufout_len) {
      puts("ERROR: output file could not be written");
      fclose(fdout);
      fclose(fdin);
      return(1);
    }

    totwrite += bufout_len;

  } while (bufin_len != 0);

  if (totread == 0) {
    puts("ERROR: input file appears to be empty (compressed to nothing)");
  } else {
    printf("%s -> %s [%zu%%] (%zu -> %zu)\n", argv[1], argv[2], (totwrite * 100) / totread, totread, totwrite);
  }

  fclose(fdout);
  fclose(fdin);
  return(0);
}
