/*
 * MVCOMP minimum overlapping depack size determination
 * Copyright (C) 2024 E. C. Masloch
 *
 * MIT License.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <inttypes.h>

#define TRANSFERBUFFERSIZE 1024
static uint8_t transferbuffer[TRANSFERBUFFERSIZE];

int readpacked(uint8_t * buffer, FILE * handle, uint32_t size) {
  uint32_t blocksize, remaining = size, read = 0;
  fseek(handle, 0, SEEK_SET);
  while (remaining) {
    blocksize = TRANSFERBUFFERSIZE;
    if (blocksize > remaining) {
      blocksize = remaining;
    }
    if (blocksize != fread(transferbuffer, 1, blocksize, handle)) {
      return 1;
    }
    memcpy(buffer + read, transferbuffer, blocksize);
    read += blocksize;
    remaining -= blocksize;
  }
  return 0;
}

int is_overlapping(uint8_t * dst, uint8_t * src) {
  return (dst > src);
}

uint16_t getword(uint8_t * buffer) {
  return buffer[0] + 256 * buffer[1];
}

void setword(uint8_t * buffer, uint16_t value) {
  buffer[0] = value & 255;
  buffer[1] = (value >> 8) & 255;
}

int decrementremaining(uint32_t * remaining, uint32_t amount) {
  if (*remaining < amount) {
    return 1;
  }
  *remaining -= amount;
  return 0;
}

int depack(uint8_t * dstbuffer, uint32_t * outsize, uint8_t * srcbuffer, uint32_t packedsize) {
  uint8_t * src = srcbuffer;
  uint8_t * dst = dstbuffer;
  uint32_t srcremaining = packedsize;
  uint32_t dstremaining = (src + srcremaining) - dst;
  uint16_t word, length, offset, repeatliteral, literalvalue;
  if (is_overlapping(dst, src)) {
    return 1;
  }
  while (srcremaining >= 2) {
    word = getword(src);
    src += 2;
    if (decrementremaining(&srcremaining, 2)) {
      return 1;
    }
    length = (word >> 12) & 15;
    if (length == 0) {
      literalvalue = word & 255;
      repeatliteral = (word >> 8) & 15;
      if (decrementremaining(&dstremaining, 1 + repeatliteral * 2)) {
        return 1;
      }
      *dst++ = literalvalue;
      while (repeatliteral) {
        setword(dst, getword(src));
        src += 2;
        if (decrementremaining(&srcremaining, 2)) {
          return 1;
        }
        dst += 2;
        repeatliteral -= 1;
      }
    } else {
      offset = word & 0xFFF;
      length += 1;
      offset += 1;
      if (decrementremaining(&dstremaining, length)) {
        return 1;
      }
      while (length) {
        *dst = *(dst - offset);
        dst += 1;
        length -= 1;
      }
      if (is_overlapping(dst, src)) {
        return 1;
      }
    }
  }
  if (srcremaining) {
    return 1;
  }
  *outsize = dst - dstbuffer;
  return 0;
}

int comparedepacked(uint8_t * buffer, FILE * handle, uint32_t size) {
  uint32_t blocksize, remaining = size, read = 0;
  fseek(handle, 0, SEEK_SET);
  while (remaining) {
    blocksize = TRANSFERBUFFERSIZE;
    if (blocksize > remaining) {
      blocksize = remaining;
    }
    if (blocksize != fread(transferbuffer, 1, blocksize, handle)) {
      return 1;
    }
    if (memcmp(buffer + read, transferbuffer, blocksize)) {
      return 1;
    }
    read += blocksize;
    remaining -= blocksize;
  }
  return 0;
}

int main(int argc, char ** argv) {
  FILE * originalfile;
  FILE * packedfile;
  uint32_t originalsize, packedsize, originalsizerounded, packedsizerounded;
  uint32_t depackbuffersize, depackedsize;
  uint8_t * depackbuffer = NULL;
  if (argc != 3) {
    printf("Usage: mvsize original.bin packed.mv\n");
    return 1;
  }
  originalfile = fopen(argv[1], "rb");
  if (! originalfile) {
    fprintf(stderr, "Error: Unable to open original file \"%s\" for reading\n", argv[1]);
    return 2;
  }
  packedfile = fopen(argv[2], "rb");
  if (! packedfile) {
    fprintf(stderr, "Error: Unable to open packed file \"%s\" for reading\n", argv[2]);
    return 2;
  }
  fseek(originalfile, 0, SEEK_END);
  originalsize = ftell(originalfile);
  if (originalsize > 0xFFFFfff0) {
    fprintf(stderr, "Error: Found original file \"%s\" too large\n", argv[1]);
    return 3;
  }
  originalsizerounded = (originalsize + 15) & ~15;
  fseek(originalfile, 0, SEEK_SET);
  fseek(packedfile, 0, SEEK_END);
  packedsize = ftell(packedfile);
  if (packedsize > 0xFFFFfff0) {
    fprintf(stderr, "Error: Found packed file \"%s\" too large\n", argv[2]);
    return 4;
  }
  packedsizerounded = (packedsize + 15) & ~15;
  fseek(packedfile, 0, SEEK_SET);
  depackbuffersize = originalsizerounded + packedsizerounded;
  if (packedsizerounded > depackbuffersize) {
    fprintf(stderr, "Error: Found files too large\n");
    return 5;
  }
  depackbuffer = malloc(depackbuffersize);
  if (! depackbuffer) {
    fprintf(stderr, "Error: Out of memory\n");
    return 255;
  }
  if (readpacked(&depackbuffer[originalsizerounded], packedfile, packedsize)) {
    fprintf(stderr, "Error: Failed to read file\n");
    return 6;
  }
  if (depack(depackbuffer, &depackedsize, &depackbuffer[originalsizerounded], packedsize)) {
    fprintf(stderr, "Error: Failed to depack with full buffer, depack fail\n");
    return 7;
  }
  if (depackedsize != originalsize) {
    fprintf(stderr, "Error: Failed to depack with full buffer, size mismatch\n");
    return 8;
  }
  if (comparedepacked(depackbuffer, originalfile, originalsize)) {
    fprintf(stderr, "Error: Failed to depack with full buffer, data mismatch\n");
    return 9;
  }
  fprintf(stderr, "Done depack with full buffer.\nS");
  fflush(stderr);
  {
    /* do a binary search in a loop. based on inicomp's test mode A.
       https://hg.pushbx.org/ecm/inicomp/file/0b1c60bfe1df/inicomp.asm#l2167 */
    uint32_t upperbound = originalsizerounded / 16;
    uint32_t lowerbound = 0;
    uint32_t dx;
    uint32_t attempts = 1;
    for (;;) {
      dx = upperbound;
      dx -= lowerbound;
      if (dx == 0) {
        break;
      }
      dx >>= 1;
      /* Rounding down, so that we never retry upper bound.
         The upper bound is known to be working.
      */
      dx += lowerbound;
      attempts += 1;
      if (readpacked(&depackbuffer[dx * 16], packedfile, packedsize)) {
        fprintf(stderr, "\nError: Failed to read file subsequently\n");
        return 10;
      }
      if (depack(depackbuffer, &depackedsize, &depackbuffer[dx * 16], packedsize)) {
        fprintf(stderr, "F");
        fflush(stderr);
        /* error: this attempt is one below the new lower bound */
        dx += 1;
        lowerbound = dx;
        continue;
      } else {
        if (depackedsize != originalsize) {
          fprintf(stderr, "\nError: Failed to depack subsequently, size mismatch\n");
          return 11;
        }
        if (comparedepacked(depackbuffer, originalfile, originalsize)) {
          fprintf(stderr, "\nError: Failed to depack subsequently, data mismatch\n");
          return 12;
        }
        fprintf(stderr, "S");
        fflush(stderr);
        /* success: this attempt is the new upper bound */
        upperbound = dx;
        continue;
      }
    }
    fprintf(stderr, "\nDone binary search, "
      "%"PRIu32" attempts, needs %"PRIu32" paragraphs before source.\n",
      attempts, upperbound);
    printf("%"PRIu32"\n", upperbound);
  }
  return 0;
}
